/*
 * This is an example of the Tcl/C++ interface.
 *
 * We define a C++ class "fraction", a numeric data type to represent
 * fractional numbers, with the numerator and denominator both a long.
 *
 * On the Tcl level, we implement the functions "add", "sub", "div"
 * and "mult", using the class's operators
 *
 * If HAVE_TCLOBJ is not defined, this class will just be a C++ class
 * like any other.
 */

#ifndef __FRACTION_H__
#define __FRACTION_H__

#ifdef HAVE_TCLOBJ
#include <tcl.h>
#include <tclobj.h>
#endif

/*
 * If we want to coexist on the Tcl level, we must use
 * Tcl_Object as our base class
 */

#ifdef HAVE_TCLOBJ
class fraction : public Tcl_Object
#else
class fraction
#endif
{
public:
  fraction (void);			/* initialize to zero          */
  fraction (long);			/* initialize to integer value */
  fraction (long, long);		/* initialize to fraction a/b  */
  fraction (const fraction &);		/* copy constructor            */

  /*
   * Assignment
   */

  fraction & operator=  (const fraction &);
  fraction & operator+= (const fraction &);
  fraction & operator-= (const fraction &);
  fraction & operator*= (const fraction &);
  fraction & operator/= (const fraction &);

  /*
   * Binary operators
   */

  fraction   operator+  (const fraction &) const;
  fraction   operator-  (const fraction &) const;
  fraction   operator*  (const fraction &) const;
  fraction   operator/  (const fraction &) const;

  /*
   * Unary operators
   */

  fraction   operator+  () const;
  fraction   operator-  () const;


  /*
   * comparison operators
   */

  int        operator<  (const fraction &) const;
  int        operator>  (const fraction &) const;
  int        operator>= (const fraction &) const;
  int        operator<= (const fraction &) const;
  int        operator== (const fraction &) const;
  int        operator!= (const fraction &) const;

  /*
   * conversion to double
   */

  operator double() const;

  /*
   * conversion to char* for output
   */

  operator const char *() const;

  /*
   * Tcl interface.
   */
#ifdef HAVE_TCLOBJ
  fraction     (Tcl_Interp *, int, char *[], int);
  int Tcl_Proc (Tcl_Interp *, int, char *[], int);
  TCL_OBJECT_DECL_CLASS(fraction);
#endif

private:
  fraction & gcf (void);	/* divide by greatest common factor */
  long num, den;
};

/*
 * Inline functions for the more trivial stuff
 */

inline
fraction::fraction (void)
{
  num = 0;
  den = 1;
}

inline
fraction::fraction (long thenum)
{
  num = thenum;
  den = 1;
}

inline
fraction::fraction (long thenum, long theden)
{
  num = thenum;
  den = theden;
}

inline
fraction::fraction (const fraction &other)
{
  num = other.num;
  den = other.den;
}

/*
 * Not very optimized, but so be it ...
 */

inline fraction
fraction::operator+ () const
{
  return *this;
}

inline fraction
fraction::operator- () const
{
  return fraction(-num,den);
}

inline fraction &
fraction::operator= (const fraction &other)
{
  num = other.num;
  den = other.den;
  return *this;
}

inline fraction &
fraction::operator-= (const fraction &other)
{
  return operator+=(-other);
}

inline fraction &
fraction::operator/= (const fraction &other)
{
  return operator*=(fraction(other.den,other.num));
}

inline fraction
fraction::operator+ (const fraction &other) const
{
  fraction temp(*this);
  return (temp+=other);
}

inline fraction
fraction::operator- (const fraction &other) const
{
  fraction temp(*this);
  return (temp-=other);
}

inline fraction
fraction::operator* (const fraction &other) const
{
  fraction temp(*this);
  return (temp*=other);
}

inline fraction
fraction::operator/ (const fraction &other) const
{
  fraction temp(*this);
  return (temp/=other);
}

inline int
fraction::operator== (const fraction &other) const
{
  return (num==other.num && den==other.den);
}

inline int
fraction::operator!= (const fraction &other) const
{
  return (!operator==(other));
}

inline int
fraction::operator<= (const fraction &other) const
{
  return (operator<(other) || operator==(other));
}

inline int
fraction::operator> (const fraction &other) const
{
  return (!operator<=(other));
}

inline int
fraction::operator>= (const fraction &other) const
{
  return (!operator<(other));
}

#endif
